import numpy as np
import matplotlib.pyplot as plt
import matplotlib 
import statannot
from sklearn.metrics import log_loss

from SMdRQA.RQA_functions import doanes_formula
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import entropy

import os
import pickle
import pandas as pd
import seaborn as sns
from random import sample
import scipy.stats as ss
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectFromModel
from sklearn.feature_selection import RFE
from sklearn.metrics import auc
from sklearn.metrics import accuracy_score
from sklearn.metrics import log_loss
from sklearn.preprocessing import normalize
import matplotlib.pyplot as plt
import itertools
from tqdm import tqdm

def HistogramSampling(histogram,bins_midpoints, samples):
  cdf=np.cumsum(histogram)
  cdf = cdf / cdf[-1]
  values = np.random.rand(samples)
  value_bins = np.searchsorted(cdf, values)
  #print('values:',values)
  #print('value bins:',value_bins)
  random_from_cdf = bins_midpoints[value_bins]
  edges=list(bins_midpoints-0.5)
  edges.append(bins_midpoints[-1]+0.5)

  hist2,edges1=np.histogram(random_from_cdf,bins=edges)
  #print('bin_mids:',(edges1[:-1]+edges1[1:])/2)
  return random_from_cdf,hist2



def Sliding_windowBS_sub_per_lam(RP, maxsize, winsize,n_boot):  #for percenthage laminarity
  num_windows=maxsize-winsize+1
  P=[]
  P_bar=[]
  for i in range(num_windows):
    sub_window=RP[i:i+winsize,i:i+winsize]
    diag_his=vert_hist(sub_window,winsize)
    #diag_his=diag_his[1:]
    P.append(diag_his/np.sum(diag_his))
    P_bar.append(diag_his)
  P=np.array(P)
  n_bar=int(np.sum(P_bar)/num_windows)
  print('n_bar:',n_bar)
  #print(np.sum(P, axis=1))
  P_summed=np.sum(P,axis=0)
  P_summed=P_summed/np.sum(P_summed)
  #print('n_bar:',n_bar)
  bins_midpoints=np.array(range(1,len(P_summed)+1))
  #line_array=IND_array+1
  VAR=[]
  for j in range(n_boot):
    hist_sample,hist=HistogramSampling(P_summed,bins_midpoints, n_bar)
    hist=hist/np.sum(hist)
    #percent_more=entropy(hist,2, winsize)
    percent_more=np.sum(hist[1:])/np.sum(hist)
    #avg_line=entropy(sel_P,1,len(sel_P)-1)
    #print(avg_line)
    #print(avg_line)
    VAR.append(percent_more)
  #return np.std(VAR)**2
  return np.quantile(VAR,0.95)-np.quantile(VAR,0.05)
  
def Sliding_windowBS_sub_avg_vert(RP, maxsize, winsize,n_boot):
  num_windows=maxsize-winsize+1
  P=[]
  P_bar=[]
  for i in range(num_windows):
    sub_window=RP[i:i+winsize,i:i+winsize]
    diag_his=vert_hist(sub_window,winsize)
    #diag_his=diag_his[1:]
    P.append(diag_his/np.sum(diag_his))
    P_bar.append(diag_his)
  P=np.array(P)
  n_bar=int(np.sum(np.mean(P_bar,axis=0)))
  print('n_bar:',n_bar)
  #print(np.sum(P, axis=1))
  P_summed=np.sum(P,axis=0)
  P_summed=P_summed/np.sum(P_summed)
  #print('n_bar:',n_bar)
  bins_midpoints=np.array(range(1,len(P_summed)+1))
  #line_array=IND_array+1
  VAR=[]
  for j in range(n_boot):
    hist_sample,hist=HistogramSampling(P_summed,bins_midpoints, n_bar)
    hist=hist/np.sum(hist)
    #percent_more=entropy(hist,2, winsize)
    percent_more=np.mean(hist_sample[1:])
    #avg_line=entropy(sel_P,1,len(sel_P)-1)
    #print(avg_line)
    #print(avg_line)
    VAR.append(percent_more)
  #return np.std(VAR)**2
  return np.quantile(VAR,0.95)-np.quantile(VAR,0.05)

def Sliding_windowBS_sub_percent_det(RP, maxsize, winsize,n_boot):
  num_windows=maxsize-winsize+1
  P=[]
  P_bar=[]
  for i in range(num_windows):
    sub_window=RP[i:i+winsize,i:i+winsize]
    diag_his=diaghist(sub_window,winsize)
    #diag_his=diag_his[1:]
    P.append(diag_his/np.sum(diag_his))
    P_bar.append(diag_his)
  P=np.array(P)
  n_bar=int(np.sum(P_bar)/num_windows)
  print('n_bar:',n_bar)
  #print(np.sum(P, axis=1))
  P_summed=np.sum(P,axis=0)
  P_summed=P_summed/np.sum(P_summed)
  #print('n_bar:',n_bar)
  bins_midpoints=np.array(range(1,len(P_summed)+1))
  #line_array=IND_array+1
  VAR=[]
  for j in range(n_boot):
    hist_sample,hist=HistogramSampling(P_summed,bins_midpoints, n_bar)
    hist=hist/np.sum(hist)
    #percent_more=entropy(hist,2, winsize)
    percent_more=np.sum(hist[1:])/np.sum(hist)
    #avg_line=entropy(sel_P,1,len(sel_P)-1)
    #print(avg_line)
    #print(avg_line)
    VAR.append(percent_more)
  #return np.std(VAR)**2
  return np.quantile(VAR,0.95)-np.quantile(VAR,0.05)
  
def Sliding_windowBS_sub_avg_diag(RP, maxsize, winsize,n_boot):
  num_windows=maxsize-winsize+1
  P=[]
  P_bar=[]
  for i in range(num_windows):
    sub_window=RP[i:i+winsize,i:i+winsize]
    diag_his=diaghist(sub_window,winsize)
    #diag_his=diag_his[1:]
    P.append(diag_his/np.sum(diag_his))
    P_bar.append(diag_his)
  P=np.array(P)
  n_bar=int(np.sum(np.mean(P_bar,axis=0)))
  print('n_bar:',n_bar)
  #print(np.sum(P, axis=1))
  P_summed=np.sum(P,axis=0)
  P_summed=P_summed/np.sum(P_summed)
  #print('n_bar:',n_bar)
  bins_midpoints=np.array(range(1,len(P_summed)+1))
  #line_array=IND_array+1
  VAR=[]
  for j in range(n_boot):
    hist_sample,hist=HistogramSampling(P_summed,bins_midpoints, n_bar)
    hist=hist/np.sum(hist)
    #percent_more=entropy(hist,2, winsize)
    percent_more=np.mean(hist_sample[1:])
    #avg_line=entropy(sel_P,1,len(sel_P)-1)
    #print(avg_line)
    #print(avg_line)
    VAR.append(percent_more)
  #return np.std(VAR)**2
  return np.quantile(VAR,0.95)-np.quantile(VAR,0.05)
    
def Sliding_windowBS(RP,maxsize,var):
  WINSIZE=[]
  VAR=[]
  for i in tqdm(range(20,maxsize,10)):
    if var == 'percent_lam':
      var_std=Sliding_windowBS_sub_per_lam(RP, maxsize, i,1000)
    elif var == 'percent_det':
      var_std=Sliding_windowBS_sub_percent_det(RP, maxsize, i,1000)
    elif var == 'avg_vert':
      var_std=Sliding_windowBS_sub_avg_vert(RP, maxsize, i,1000)
    elif var == 'avg_diag':
      var_std=Sliding_windowBS_sub_avg_diag(RP, maxsize, i,1000)
    
    WINSIZE.append(i)
    VAR.append(var_std)

  DICT={'WINSIZE':WINSIZE,'95% quantile- 5% quantile':VAR}
  df=pd.DataFrame.from_dict(DICT)
  return df
  
def Sliding_window_whole_data(RP_dir, var):
  files = os.listdir(RP_dir)
  DFs = []
  for File in files:
    path = RP_dir+'/'+File
    info = File.split('-')[0]
    RP = np.load(path)
    maxsize = len(RP)
    df = Sliding_windowBS(RP,maxsize,var)
    df['group'] = info
    DFs.append(df)
    
  df_out = pd.concat(DFs)
  
  return df_out
    
    



import pandas as pd
import numpy as np
from itertools import combinations
from sklearn.model_selection import RepeatedStratifiedKFold
from sklearn.neighbors import KNeighborsClassifier
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.preprocessing import StandardScaler

def feature_selection(x_train, y_train, features, y_name, inner_splits, select_k=False, repeats=5): # inner loop function
    '''
    Function used for feature selection in nested CV, using best subset setection and average cross validation accuracy
    Input:____________________________________________________________________________________________________________________________________________
    x_train                   : training set from the outer loops of nested cross validation function
    y_train                   : output variables for the training set from the outer loops of the nested cross validation
    features                  : total set of features from which features are being selected
    y_name                    : column name for the outcome variable
    inner_splits              : cross validation splits for feature selection
    select_k                  : whether the user wants to tune k also based on training set, default = False, k=5
    repeats                   : number of repeats for cross validation for feature selection
    Output:_____________________________________________________________________________________________________________________________________________
    best_features             : list of features got selected
    best_score                : avg accuracy for the best feature subset
    best_roc_auc              : avg ROC AUC for the best feature subset
    '''
    from tqdm import tqdm
    # Load the data into a pandas DataFrame
    
    if select_k == False: 
      X=x_train
      Y=y_train
      # Split the data into features and outcome variable=

      # Define the number of features to consider in each subset
      k_values = range(1, len(X.columns) + 1)

      # Define the cross-validation scheme
      cv = RepeatedStratifiedKFold(n_splits=inner_splits, n_repeats=repeats,random_state=1)

      # Initialize variables to store the best feature set and its score
      best_score = 0
      best_roc_auc=0
      best_features = None

      # Iterate over different feature sets and evaluate their performance
      for k in tqdm(k_values):
          for feature_set in combinations(X.columns, k):
              X_sel = X[list(feature_set)]

              # Evaluate the score using cross-validation
              scores = []
              roc_aucs=[]
              for train_idx, test_idx in cv.split(X_sel, Y):
                  X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
                  y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]

                  knn = KNeighborsClassifier()
                  knn.fit(X_train, y_train)
                  score = accuracy_score(y_test, knn.predict(X_test))
                  roc_auc = roc_auc_score(y_test, knn.predict_proba(X_test)[:, 1])
                  scores.append(score)
                  roc_aucs.append(roc_auc)

              # Calculate the mean score across cross-validation repeats
              mean_score = sum(scores) / len(scores)
              mean_roc_auc= sum(roc_aucs) / len(roc_aucs)

              # Update the best feature set if necessary
              if mean_score > best_score:
                  best_score = mean_score
                  best_features = feature_set
                  best_roc_auc= mean_roc_auc

      return list(best_features), best_score, best_roc_auc
      
    elif select_k == True:
      X=x_train
      Y=y_train
      k_max= int(len(Y)/2)
      
      
             
      
      k_values = range(1, len(X.columns) + 1)

      # Define the cross-validation scheme
      cv = RepeatedStratifiedKFold(n_splits=inner_splits, n_repeats=repeats,random_state=1)

      # Initialize variables to store the best feature set and its score
      best_score = 0
      best_roc_auc=0
      best_features = None

      # Iterate over different feature sets and evaluate their performance
      for k in tqdm(k_values):
          for feature_set in combinations(X.columns, k):
              X_sel = X[list(feature_set)]

              # Evaluate the score using cross-validation
              scores = []
              roc_aucs=[]
              for train_idx, test_idx in cv.split(X_sel, Y):
                  X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
                  y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]

                  knn = KNeighborsClassifier()
                  knn.fit(X_train, y_train)
                  score = accuracy_score(y_test, knn.predict(X_test))
                  roc_auc = roc_auc_score(y_test, knn.predict_proba(X_test)[:, 1])
                  scores.append(score)
                  roc_aucs.append(roc_auc)

              # Calculate the mean score across cross-validation repeats
              mean_score = sum(scores) / len(scores)
              mean_roc_auc= sum(roc_aucs) / len(roc_aucs)

              # Update the best feature set if necessary
              if mean_score > best_score:
                  best_score = mean_score
                  best_features = feature_set
                  best_roc_auc= mean_roc_auc
      
      
      neighbour_values = range(1, k_max)
      
      cv1 = RepeatedStratifiedKFold(n_splits=inner_splits, n_repeats=5*repeats,random_state=1)
      
      best_score1=0
      best_roc_suc1=0
      best_neigh= None
      X_sel = X[list(best_features)]
      for neigh in tqdm(neighbour_values):
          scores1=[]
          roc_aucs1=[]
          for train_idx, test_idx in cv1.split(X, Y):
              X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
              y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]
              knn = KNeighborsClassifier(n_neighbors=neigh)
              knn.fit(X_train, y_train)
              score = accuracy_score(y_test, knn.predict(X_test))
              roc_auc = roc_auc_score(y_test, knn.predict_proba(X_test)[:, 1])
              scores1.append(score)
              roc_aucs1.append(roc_auc)
              
          mean_score1 = sum(scores1) / len(scores1)
          mean_roc_auc1= sum(roc_aucs1) / len(roc_aucs1)
          
          if mean_score1 > best_score1:
             best_score1 = mean_score1
             best_neigh = neigh
             best_roc_auc1= mean_roc_auc1 
             
          
          
      #best_neigh = max(best_neigh, 5) 
                
      return list(best_features), best_score, best_roc_auc, best_neigh
          
      
      

def nested_cv(data_file, features, y_name, outname, repeats=1000, inner_repeats=10, outer_splits=3, inner_splits=2):
    '''
    This is a function to run nested cross validation. The outer loop is for evaluation and inner loop for feature selection
    Input:__________________________________________________________________________________________________________________
    data_file             : pandas dataframe 
    features              : features that should be used (column names)
    y_name                : outcome column name
    outname               : Name that should be added to the output file
    repeats               : number of times the outer loops should repeat
    inner_repeat          : number of times the inner loop should repeat
    outer split           : number of cross validation splits for the outer loop
    inner_split           : number of cross validation splits for the inner loop
    Output:___________________________________________________________________________________________________________________
    CSV File              : containing cross validation accuracy and ROC AUC for each of the outer loops
    '''
    from tqdm import tqdm
    

      
    # Load the data into a pandas DataFrame
    df = data_file

    # Split the data into features and outcome variable
    X = df[features]
    y = df[y_name]

    # Define the cross-validation scheme for the outer loop
    outer_cv = RepeatedStratifiedKFold(n_splits=outer_splits, n_repeats=repeats,random_state=1)

    # Initialize variables to store the best feature set and its score
    best_score = 0
    best_features = None
    FEATURES=[]
    ACCURACY=[]
    ROC_AUC=[]
    val_ACC=[]
    val_ROC_AUC=[]
    # Iterate over the outer cross-validation splits
    for train_idx, test_idx in tqdm(outer_cv.split(X, y)):
        X_train, X_test = X.iloc[train_idx], X.iloc[test_idx]
        y_train, y_test = y.iloc[train_idx], y.iloc[test_idx]

        # Define the cross-validation scheme for the inner loop
        best_features, best_score, best_roc_auc, best_neigh = feature_selection(X_train, y_train, features, y_name, inner_splits,select_k=True, repeats=inner_repeats)
        val_ACC.append(best_score)
        val_ROC_AUC.append(best_roc_auc)
        FEATURES.append(best_features)
        print('selected number of neighbours is:', best_neigh)
        knn = KNeighborsClassifier(n_neighbors=best_neigh)
        knn.fit(X_train[best_features], y_train)
        score = accuracy_score(y_test, knn.predict(X_test[best_features]))
        ACCURACY.append(score)
        print('accuracy:', score)
        roc_auc = roc_auc_score(y_test, knn.predict_proba(X_test[best_features])[:, 1])
        ROC_AUC.append(roc_auc)

    
    DF_out=pd.DataFrame.from_dict({'features':FEATURES, 'validation_ROC_AUC':val_ROC_AUC, 'validation_accuracy':val_ACC,'test_accuracy':ACCURACY, 'test_ROC_AUC':ROC_AUC})
    DF_out.to_csv('Nested_CV_'+outname+'_fin_results.csv')





import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import os
import numpy as np
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import pickle
import random
from scipy.stats import skew
from p_tqdm import p_map
from functools import partial
from scipy.interpolate import pchip_interpolate
import memory_profiler

def doanes_formula(data,n):
    '''
    Functions for computing number of bins for data using Doane's formula
    data : input array
    n : number of elements in data
    '''
    
    g1 = skew(data)
    sigma_g1 = np.sqrt((6 * (n - 2)) / ((n + 1) * (n + 3)))

    k = 1 + np.log2(n) + np.log2(1 + np.abs(g1) / sigma_g1)
    
    for i in range(len(k)):
      k[i] = int(np.ceil((2/3)*k[i]))  # Round up to the nearest integer

    return k
    
def binscalc(X,n,d, method):
    """
    Function to calculate number of bins in a histogram using a generalised Freedman–Diaconis rule.
    
    Arguments:  X    double array of shape (n,d)    time series. Think of it as n points in a d dimensional space
                n    int                            number of samples in time series
                d    int                            number of measurements at each time step
                
    Output:          int array of shape (d)         Number of bins along each axis
    """
    if method == 'FD': #generalised Freedman–Diaconis rule.
      mult_fact=(n**(1/(2+d+10**(-9)))) # Generalised Freedman-Diaconis rule with Bins[i] \propto n^(1/(d+2))
      Bins=np.zeros(d)
      inf_arr=np.amin(X,axis=0)
      sup_arr=np.amax(X,axis=0)
      q25=np.quantile(X,0.25,axis=0)
      q75=np.quantile(X,0.75,axis=0)
      IQR=(q75-q25)+10**(-9)
      print('IQR:',IQR)
      Bins_1=np.ceil(mult_fact*(sup_arr-inf_arr)/(2*IQR))
      Bins=Bins_1.astype(int)
    elif method== 'sqrt':
      ns=[n]*d
      Bins_1=np.ceil(np.sqrt(ns))
      Bins=Bins_1.astype(int)
      
    elif method == 'rice':
      ns=[n]*d
      Bins_1=2*np.ceil(np.cbrt(ns))
      Bins=Bins_1.astype(int)
    elif method == 'sturges':
      ns=[n]*d
      Bins_1=1+np.ceil(np.log(ns))
      Bins=Bins_1.astype(int)
    elif method== 'doanes':
      Bins_1=doanes_formula(X,n)
      Bins=Bins_1.astype(int)
      
    elif method == 'default':
      Bins=15
      
    
    #print('inf:',inf)
    #print('sup:',sup)
    #print('IQR:',iqr)
     
    return Bins
    

def mutualinfo(X,Y,n,d):
    """
    Function to calculate mutual information between to time series
    
    Arguments:  X,Y  double array of shape (n,d)    time serieses
                n    int                            number of samples in time series
                d    int                            number of measurements at each time step
                
    Output:          double                         mutual information between X,Y
    """
    points=np.concatenate((X,Y),axis=1)
    bins=binscalc(points,n,2*d,'FD')   
    print('BINS:', bins)    
    ###print('bins:',bins.shape)
    ###print('points:',points.shape)
    p_xy=np.histogramdd(points,bins=binscalc(points,n,2*d,'default'))[0]+10**(-9) # 10^-9 added so that x log x does not diverge when x=0 in the calculation of mutual information
    p_x=np.histogramdd(X,bins=binscalc(X,n,d,'default'))[0]+10**(-9)
    p_y=np.histogramdd(Y,bins=binscalc(Y,n,d,'default'))[0]+10**(-9)
    p_xy/=np.sum(p_xy) # Normalising the probability distribution
    p_x/=np.sum(p_x)
    p_y/=np.sum(p_y)
    return np.sum(p_xy*np.log2(p_xy))-np.sum(p_x*np.log2(p_x))-np.sum(p_y*np.log2(p_y)) # formula for mutual information

def timedelayMI(u,n,d,tau):
    """
    Function to calculate mutual information between a time series and a delayed version of itself
    
    Arguments:  u    double array of shape (n,d)    time series
                n    int                            number of samples in time series
                d    int                            number of measurements at each time step
                tau  double                         amount of delay
                
    Output:          double                         mutual information between u and u delayed by tau
    """
    X=u[0:n-tau,:]
    Y=u[tau:n,:]
    return mutualinfo(X,Y,n-tau,d)
    
def findtau(u,n,d,grp):
    """
    Function to calculate correct delay for estimating embedding dimension
    
    Arguments:  u    double array of shape (n,d)    time series
                n    int                            number of samples in time series
                d    int                            number of measurements at each time step
                
    Output:          double                         delay such that the mutual information between u and u delayed by tau is at its first minimum as a function of tau
    """
    TAU=[]
    MIARR=[]
    minMI=timedelayMI(u,n,d,1)
    for tau in range(2,n):
        nextMI=timedelayMI(u,n,d,tau)
        TAU.append(tau)
        MIARR.append(nextMI)
        if nextMI>minMI:
          break
        minMI=nextMI
        
        
    
    return tau-1

#### Calculation of m ###################################################################################################################################################################

def delayseries(u,n,d,tau,m):
    """Delay series"""
    s=np.zeros((n-(m-1)*tau,m,d))
    for i in range(n-(m-1)*tau):
        for j in range(m):
            s[i,j]=u[i+j*tau]
    return s

def nearest(s,n,d,tau,m):
    nn=np.zeros(n-m*tau,dtype=int)
    nn[0]=n-m*tau-1
    for i in range(n-m*tau):
        for j in range(n-m*tau):
            if(i!=j and np.linalg.norm(s[i]-s[j])<np.linalg.norm(s[i]-s[nn[i]])):
                nn[i]=j
    return nn
                
def fnnratio(u,n,d,m,tau,r,sig):
    """Calculates the ratio of false nearest neighbours. Vary m to find stabilising trend"""
    s1=delayseries(u,n,d,tau,m)     # embedding in m dimensions
    s2=delayseries(u,n,d,tau,m+1)   # embedding in m+1 dimensions
    nn=nearest(s1,n,d,tau,m)        # containg nearest neghbours after embedding in m dimensions
    isneigh=np.zeros(n-m*tau)
    isfalse=np.zeros(n-m*tau)
    for i in range(n-m*tau):
        disto=np.linalg.norm(s1[i]-s1[nn[i]])+10**(-9)
        distp=np.linalg.norm(s2[i]-s2[nn[i]])
        if(disto<sig/r):
            isneigh[i]=1
            if(distp/disto>r):
                isfalse[i]=1
    return sum(isneigh*isfalse)/(sum(isneigh)+10**(-9))

def fnnhitszero(u,n,d,m,tau,sig,delta,Rmin,Rmax,rdiv):
    Rarr=np.linspace(Rmin,Rmax,rdiv)
    for i in range(rdiv):
        if fnnratio(u,n,d,m,tau,Rarr[i],sig)<delta:
            return Rarr[i]
    return -1
    
def findm(u,n,d,tau,sd,delta,Rmin,Rmax,rdiv,bound):
   
    mmax=int((3*d+11)/2)
    rm=fnnhitszero(u,n,d,mmax,tau,sd,delta,Rmin,Rmax,rdiv)
    rmp=fnnhitszero(u,n,d,mmax+1,tau,sd,delta,Rmin,Rmax,rdiv)
    
    if(rm-rmp>bound):
        return mmax+1
    for m in range(1,mmax):
        rmp=rm
        rm=fnnhitszero(u,n,d,mmax-m,tau,sd,delta,Rmin,Rmax,rdiv)
        print('rm-rmp:',rm-rmp)
        if(rm-rmp>bound):
            return mmax+1-m
              
          
            
    return -1 
      
    
    
### Calculation of epsilon ######################################################################################################################################################
   
def reccplot(u,n,d,m,tau,eps):
    """"Creates reccurrence plot"""
    #normarr=[]
    s=delayseries(u,n,d,tau,m) 
    rplot=np.zeros((n-(m-1)*tau,n-(m-1)*tau),dtype=int)
    for i in range(n-(m-1)*tau):
        for j in range(n-(m-1)*tau):
            #normarr.append(np.linalg.norm(s[i]-s[j]))
            if np.linalg.norm(s[i]-s[j])<eps:
                rplot[i,j]=1          
    return rplot
    
def reccrate(rplot,n):
    return float(np.sum(rplot))/(n*n)    
    
def findeps(u,n,d,m,tau,reqrr,rr_delta,epsmin,epsmax,epsdiv):
    eps=np.linspace(epsmin,epsmax,epsdiv)
    s=delayseries(u,n,d,tau,m) 
    for k in range(epsdiv):
        rplot=np.zeros((n-(m-1)*tau,n-(m-1)*tau),dtype=int)
        for i in range(n-(m-1)*tau):
            for j in range(n-(m-1)*tau):
                if np.linalg.norm(s[i]-s[j])<eps[k]:
                    rplot[i,j]=1
        rr=reccrate(rplot,n-(m-1)*tau)
        if np.abs(rr-reqrr)<rr_delta:
            return eps[k]
        
          
    return -1

### Calculation of RQA parameters #######################################################################################################################################################

def plotwindow(M,n,win,i,j):
    window=np.zeros((win,win))
    for a in range(win):
        for b in range(win):
            window[a,b]=M[i+a,j+b]
    return window

def vert_hist(M,n):   ##### Functio to calculate vertical line distribution
    nvert=np.zeros(n+1)
    for i in range(n):
        counter=0
        for j in range(n):
            if M[j][i]==1:
                counter+=1
            else:
                nvert[int(counter)]+=1
                counter=0
        nvert[counter]+=1
    return nvert

def onedhist(M,n):
    hst=np.zeros(n+1)
    counter=0
    for i in range(n):
        if M[i]==1:
            counter+=1
        else:
            hst[counter]+=1
            counter=0
    hst[counter]+=1
    return hst

def diaghist(M,n):   #Function to calculate diagonal line distribution
    dghist=np.zeros(n+1)
    for i in range(n):
        diag=np.zeros(n-i)
        for j in range(n-i):
            diag[j]=M[i+j][j]
        subdiaghist=onedhist(diag,n-i)
        for k in range(n-i+1):
            dghist[k]+=subdiaghist[k]
    dghist*=2
    dghist[n]/=2
    return dghist


    
### Measures to capture probability distributions #######################################################################################################################################

def percentmorethan(hst,mini,n):
    numer=0
    denom=10**(-7)
    for i in range(mini,n+1):
        numer+=i*hst[i]
    for i in range(1,n+1):
        denom+=i*hst[i]
    return numer/denom

def mode(hst,mini,n):
    p=mini
    for i in range(mini+1,n+1):
        if hst[i]>hst[p]:
            p=i
    return p
    
def maxi(hst,mini,n):
    lmax=1
    for i in range(1,n-1):
      if hst[i]!=0:
        lmax=i
    return lmax

def average(hst,mini,n):
    numer=0
    denom=10**(-7)
    for i in range(mini,n+1):
        numer+=i*hst[i]
        denom+=hst[i]
    return numer/denom
    
def entropy(hst,mini,n):
    summ=0
    entr=0
    for i in range(mini,n+1):
        summ+=hst[i]
    for i in range(mini,n+1):
        if(hst[i]!=0):
            entr-=(hst[i]/summ)*np.log(hst[i]/summ)
    return entr
    




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="SMdRQA",
    version="2024.03.24",
    author="Swarag Thaikkandi, K.M. Sharika, Miss Nivedita",
    author_email="tsk4at@gmail.com",
    description="A Python implementation of MdRQA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SwaragThaikkandi/Sliding_MdRQA",
    packages=find_packages(),
    install_requires=['numpy','scipy','operator-courier','contextlib2','pytest-warnings','matplotlib','pandas','tqdm','memory_profiler','kuramoto','networkx','p_tqdm'],
    setup_requires=['numpy','scipy','operator-courier','contextlib2','pytest-warnings','matplotlib','pandas','tqdm','memory_profiler','kuramoto','networkx','p_tqdm'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)




import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import os
import numpy as np
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import pickle
import random
from scipy.stats import skew
from p_tqdm import p_map
from functools import partial
from scipy.interpolate import pchip_interpolate
import memory_profiler

from SMdRQA.RQA_functions import doanes_formula
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import entropy
import scipy.stats as ss

def Mode(array):
  '''
  Function to compute mode from a continuous data
  '''
  (counts, edges)=np.histogram(array)
  avgs=(edges[1:]+edges[:-1])/2
  avg=avgs[counts==np.amax(counts,keepdims=True)[0]]
  return avg[0]
  
def Check_Int_Array(array):
  '''
  This function checks whether an array consists of integers or not
  '''
  N=len(array)
  sum=0
  for i in range(N):
    if array[i]==int(array[i]):
       sum=sum+1

  if N==sum:
    out=1
  elif N!=sum:
    out=0

  return out
  
def Sliding_window(RP, maxsize, winsize):
  '''
  This is a function that would extract RQA variables from each sliding window for a given RP and for a specified window size
  Inputs _______________________________________________________________________
  RP                            : recurrence plot(numpy 2D matrix)
  maxsize                       : maximum size of the recurrence plot
  winsize                       : size of the window we are using
  Outputs_______________________________________________________________________
  Dict                          : dictionary containing RQA variables from all sliding windows, keys are window number
  '''
  num_windows=maxsize-winsize+1
  Dict={}
  for i in range(num_windows):
    sub_dict={}
    sub_window=RP[i:i+winsize,i:i+winsize]
    sub_dict['recc_rate']=reccrate(sub_window,winsize)
    vert_his=vert_hist(sub_window,winsize)
    diag_his=diaghist(sub_window,winsize)
    sub_dict['percent_det']=percentmorethan(diag_his,2,winsize)
    sub_dict['avg_diag']=average(diag_his,2,winsize)
    sub_dict['max_diag']=maxi(diag_his,2,winsize)
    sub_dict['percent_lam']=percentmorethan(vert_his,2,winsize)
    sub_dict['avg_vert']=average(vert_his,2,winsize)
    sub_dict['vert_ent']=entropy(vert_his,2,winsize)
    sub_dict['diag_ent']=entropy(diag_his,2,winsize)
    sub_dict['vert_max']=maxi(vert_his,2,winsize)
    
    Dict[i]=sub_dict

  return Dict
  
def Whole_window(RP,maxsize):
  '''
  Inputs _______________________________________________________________________
  Function for computing the RQA variables from the whole RP
  RP                             : recurrence plot(numpy 2D matrix)
  maxsize                        : size of RP
  Outputs ______________________________________________________________________
  sub_dict                       : dictionary containing values of each of the RQA variables from the entire RP
  '''
  Dict={}
  
  sub_dict={}
  sub_dict['recc_rate']=reccrate(RP,maxsize)
  vert_his=vert_hist(RP,maxsize)
  diag_his=diaghist(RP,maxsize)
  sub_dict['percent_det']=percentmorethan(diag_his,2,maxsize)
  sub_dict['avg_diag']=average(diag_his,2,maxsize)
  sub_dict['max_diag']=maxi(diag_his,2,maxsize)
  sub_dict['percent_lam']=percentmorethan(vert_his,2,maxsize)
  sub_dict['avg_vert']=average(vert_his,2,maxsize)
  sub_dict['vert_ent']=entropy(vert_his,2,maxsize)
  sub_dict['diag_ent']=entropy(diag_his,2,maxsize)
  sub_dict['vert_max']=maxi(vert_his,2,maxsize)
    


  return sub_dict
  
def windowed_RP(winsize,RP_dir, save_path):
  '''
  Inputs _______________________________________________________________________
  This function computes RQA variable from each window and stores that in a dictionary
  type                            : Type of data, whether its GD/pre GD/randomized
  winsize                         : Size of window to be used
  RP_dir                          : directory where the RPs are stored as numpy files
  save_path                       : directory in which the dictionary should be saved in the form of a pickle file
  Output _______________________________________________________________________
  Dict                            : Dictionary containing RQA variables from the sliding windows
  '''
  dir=RP_dir
  files=os.listdir(dir)
  size_arr=[]
  Dict={}
  for file in files:
    a=file.strip().split('.')[0]  # To have file name as an identifier of the variable. Note that the file name should not contain "." other than the one before "npy"
    
    RP_s=np.load(dir+'/'+file,allow_pickle=True)
    (M,N)=RP_s.shape
    sub_Dict1=Sliding_window(RP_s, M, winsize)
    sub_dict2=Whole_window(RP_s, M)
    Dict[a]={'sliding':sub_Dict1,'Whole':sub_dict2}

  with open(save_path+'/whole_.pkl', 'wb') as handle:
    pickle.dump(Dict, handle)

  return Dict
  
def First_middle_last_avg(DICT, var, win_loc):
  '''
  This is a function to compute different summary statistics from the RQA variable distribution that we would get from the windows
  Inputs _______________________________________________________________________
  DICT_pre                          : dictionary containing RQA variables from data
  win_loc: window of interest(fist/last/mean/median etc)
  Outputs ______________________________________________________________________
  data                              : pandas dataframe containing the calculated summary statistic(specified) for the specified variable across different datasets
  '''
  Grp_IDs=list(DICT.keys())
  num_plots=len(Grp_IDs)
  vars=[]
  label=[]
  out=[]
  window=[]
  GID=[]
  for i in range(len(Grp_IDs)):
    num=Grp_IDs[i]
    grp_dyn=DICT[num]['sliding']
    sub_keys=grp_dyn.keys()
    recc_rate=[]
    wind=[]
    vars_sub4=[]
    label_sub4=[]
    for elem in sub_keys:
      

      vars_sub4.append(grp_dyn[elem][var])
      label_sub4.append(elem)

    if win_loc=='first':
      index=0
      app_var=vars_sub4[index]
    elif win_loc=='middle':
      index=int(np.ceil(len(vars_sub4)/2))
      app_var=vars_sub4[index]
      print('Length of array:',len(vars_sub4))
      print('Index of middle element:',index)
    elif win_loc=='last':
      index=-1
      app_var=vars_sub4[index]
    elif win_loc=='max':
      app_var=np.max(vars_sub4)
    elif win_loc=='avg':
      app_var=np.mean(vars_sub4)
    elif win_loc=='mode':
      Isint=Check_Int_Array(vars_sub4)
      if Isint==0:
        app_var=Mode(vars_sub4)
      elif Isint==1:
        (app_var,count)=ss.mode(vars_sub4)
    elif win_loc=='median':
      app_var=np.median(vars_sub4)
    elif win_loc=='whole':
      app_var=DICT[num]['Whole'][var]
      

      
    
    vars.append(app_var)
    label.append('Lorenz')
    window.append(win_loc)
    GID.append(num)

    

  DICT={'label':label,var:vars,'window':window,'group':GID}
  data=pd.DataFrame.from_dict(DICT)
  #sns.displot(data=data, x=var, kde=True,hue=['label','outcome'],kind='hist',norm_hist=True)
  return data
  
def First_middle_last_sliding_windows(DICT, var):
  '''
  Function to compute all summary statistics that we want for a given RQA variable
  Inputs _______________________________________________________________________
  DICT                          : dictionary containing RQA variables from data collected 
  var                               : specific RQA variable we are looking for
  Outputs ______________________________________________________________________
  full_data                         : pandas dataframe containing all summary statistics that we want for a given RQA variable
  '''
  win_locs=['avg','median','mode', 'whole']
  dfs=[]
  for win_loc in tqdm(win_locs):
    data=First_middle_last_avg(DICT,var, win_loc)
    dfs.append(data)

  full_data=pd.concat(dfs)
  a=full_data.columns[1]
  full_data=full_data.reset_index(drop=True)

  #sns.histplot(full_data,x=var)
  #g = sns.FacetGrid(full_data, col="window",  row="label")
  #g.map(sns.histplot,x=a,hue='outcome',norm_hist=True)
  return full_data
  
def First_middle_last_sliding_windows_all_vars(DICT, save_path):
  '''
  Function to compute all summary statistics that we want for all RQA variables
  Inputs _______________________________________________________________________
  DICT                          : dictionary containing RQA variables from data collected 
  Outputs ______________________________________________________________________
  data_out                         : pandas dataframe containing all summary statistics that we want for all RQA variable
  '''
  win_locs=['avg','median','mode', 'whole']
  vars=['recc_rate','percent_det','avg_diag','max_diag','percent_lam','avg_vert','vert_ent','diag_ent','vert_max']
  dfs=[]
  for var in tqdm(vars):
    df=First_middle_last_sliding_windows(DICT, var)
    dfs.append(df)

  data_out=pd.concat(dfs,axis=1)
  data_out=data_out.T.drop_duplicates().T
  #data_out.to_csv('/content/drive/MyDrive/MS20181013/NL_test/results/whole_diff_windows38.csv')
  #data_out.to_csv('whole_diff_windows.csv')
  data_out.to_csv(save_path)
  
  



#!/usr/bin/python

__docformat__ = "restructuredtext"

__version__ = "0.0.1"

import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import os
import numpy as np
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import pickle
import random
from scipy.stats import skew
from p_tqdm import p_map
from functools import partial
from scipy.interpolate import pchip_interpolate
import memory_profiler
import ast

from kuramoto import Kuramoto, plot_phase_coherence, plot_activity # see https://github.com/fabridamicelli/kuramoto
# For installing the kuramoto package run: pip install kuramoto





import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import os
import numpy as np
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import pandas as pd
from os import listdir
from os.path import isfile, join
from collections import defaultdict
import csv
from tqdm import tqdm
import pickle
import random
from scipy.stats import skew
from p_tqdm import p_map
from functools import partial
from scipy.interpolate import pchip_interpolate
import memory_profiler

from SMdRQA.RQA_functions import doanes_formula
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import entropy

def RP_computer(input_path, RP_dir,rdiv=451, Rmin=1, Rmax=10, delta=0.001, bound=0.2, reqrr=0.1, rr_delta=0.005, epsmin=0, epsmax=10, epsdiv=1001, windnumb=1):
  '''
  Input arguments_____________________________________________________________________________
  input_path       : folder containing the numpy files, rows> number of samples, columns> number of streams
  RP_dir           : directory in which the RPs should be stored
  rdiv             : number of divisions(resolution) for the variable r during parameter search for embedding dimension
  Rmin             : minimum value for the variable r during parameter search for embedding dimension
  Rmax             : maximum value for the variable r during parameter search for embedding dimension
  delta            : the tolerance value below which an FNN value will be considered as zero
  bound            : This is the value in the r value(at which FNN hits zero) va embedding dimension plot. The search is terminated if the value goes below this tolerance value and the value just below tolerance value is reported for embedding dimmension
  req_rr           : This is a variable that user can define. This controls the overall recurrence rate of the whole RP
  rr_delta         : this variable is used to define tolerance for accepting a value of neighbourhood radius. If the absolte differenece between the resccurence rate value to that of the desired value is less than this tolerance, the value of epsilon is accepted
  eps_min          : Minimum value of the neighbourhood radius value to being with for fixing the reccurrence rate
  eps_max          : Maximum value of the neighbourhood radius value above which the search won't progress
  eps_div          : Number of divisions between eps_min and eps_max
  windnumb         : number of windows getting extracted
  Output_______________________________________________________________________________________
  
  RPs              : recurrence plots saved as npy files in the given directory
  Error_Report_Sheet : Analogous to a log file, will record those instances where RP computation failed either due to memory error or due to value error
  param_Sheet      : RQA parameter estimated for each files
  '''
  path=input_path
  files=os.listdir(path)
  ERROROCC=[]
  FILE=[]
  TAU=[]
  Marr=[]
  EPS=[]
  BOUND=[]
  
  for File in tqdm(files):
    try:
      file_path=path+'/'+File
      data=np.load(file_path)
      (M,N)=data.shape
    
      data = (data - np.mean(data, axis=0, keepdims=True))/np.std(data, axis=0, keepdims=True)
        
      
      n=M
      d=N
      u=data
      
      sd=3*np.std(u)
      print('starting tau calculation ...')
      tau=findtau(u,n,d,0)
      print('Done Tau calculation....')
      print('TAU:',tau)
      print('starting m calculation ...')
      #notFound = 1
      #while notFound == 1: 
        #try: 
      try:
        m=findm(u,n,d,tau,sd,delta,Rmin,Rmax,rdiv,bound)
        print('Done m calculation....')
        print('m:',m)
        print('starting eps calculation ...')
        eps=findeps(u,n,d,m,tau,reqrr,rr_delta,epsmin,epsmax,epsdiv)
        print('Done eps calculation....')
        print('EPS:',eps)
        rplot=reccplot(u,n,d,m,tau,eps)
        print('Done rplot calculation....')
        rplotwind=rplot
        np.save(RP_dir+'/'+File,rplotwind)
        
          #notFound = 0
        FILE.append(File)
        TAU.append(tau)
        Marr.append(m)
        EPS.append(eps)
        BOUND.append(bound)
        
      except ValueError:
        print('unable to compute due to value error')
        ERROROCC.append(File)
       
            
    except MemoryError:
      print("Couldn't do computation due to numpy.core._exceptions.MemoryError")
      ERROROCC.append(File)

    
  DICT={'error occurances': ERROROCC}
  df_out=pd.DataFrame.from_dict(DICT)
  df_out.to_csv('Error_Report_Sheet.csv')

  DICT2={'file':FILE,'tau':TAU,'m':Marr,'eps':EPS,'bound':BOUND}
  df_out2=pd.DataFrame.from_dict(DICT2)
  df_out2.to_csv('param_Sheet.csv')
